﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.IO;
using System.Xml.Linq;
using System.Diagnostics;
using System.Reflection;

namespace XMLconfigLib
{
    /// <summary>
    /// This class converts an XML file into a dictionary<string, string> (keyword, value).
    /// E.G. of an XML file
    /// <code>
        
    /// </code>
    /// </summary>
  
    public class XConfig: IDisposable
    {
        public string fileName;
        private Dictionary<string, string> keywordConfig = new Dictionary<string, string>(10);
        public static XConfig theConfig = null;  //should only be one per executable, 
                                                 //and not deleted until it is disposed,
                                                //whereupon it is saved.
        public bool newFile = false;  //if true, old config not found, new one was set up this run.
        public Dictionary<string, string> initValues = new Dictionary<string, string>(10);


        /*  The dictionary parameter is the initial values.  They should be strings, or convertable to strings
            in a way that the string representation can be converted back via if (xxx.TryParse(out ok)....
            call constructor like:

            XConfig config = new XConfig(
                new Dictionary<string, object>
            {
                ["doChime"] = true,
                ["doSpeak"] = false,
                ["speakText"] = "time to stretch",
                ["timeout"] = 30,
            }) ;
        */

        public XConfig(string fileName, Dictionary<string, object> initValues)
        {
            this.fileName = fileName;
            if (initValues == null)
                initValues = new Dictionary<string, object>(20);
            foreach (KeyValuePair<string, object> kv in initValues)
                this.initValues[kv.Key] = kv.Value.ToString();
            theConfig = this;
            init();
        }

        public XConfig(Dictionary<string, object> initValues)
            : this(Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location)
             + Path.DirectorySeparatorChar + "xconfig.xml", initValues)
        {
            string aFile = Assembly.GetExecutingAssembly().Location;
            string productVersion = FileVersionInfo.GetVersionInfo(aFile).ProductVersion;
            if (initValues == null)
                initValues = new Dictionary<string, object>(20);
        }

        public string addInitValue(string key, object initVal)
        {
            string inVal = initVal.ToString();
            theConfig.initValues[key] = inVal;
            return this[key];
            
        }

        public void addInitValues(Dictionary<string, object> initVals)
        {
            if (initVals == null)
                return;
            foreach (KeyValuePair<string, object> kv in initVals)
            {
                theConfig.initValues[kv.Key] = kv.Value.ToString();
            }
        }

        public static List<string> asList(string input)  //allow list of things as a target
        {
                //assume members are separated by a comma
            string[] stuff = input.Split(new char[] { ',' }, StringSplitOptions.RemoveEmptyEntries);
            List<string> output = new List<string>(stuff.Length + 1);
            foreach (string s in stuff)
            {
                output.Add(s);          
            }

            return output;
        }


        /// <summary>
        /// Get or set a value in the dictionary.  Return null on get when key is not found.  Set converts
        /// input to a string; get returns an object for you to cast as necessary.
        /// </summary>
        /// <param name="key"></param>
        /// <returns></returns>
        public string this[string key]
        {
            get { 
                string outVal;
                if (keywordConfig.TryGetValue(key, out  outVal))
                    return outVal;
                if (initValues.TryGetValue(key, out outVal))
                    return outVal;
                throw new ApplicationException("key not in dictionary: " + key);
            }
            set { keywordConfig[key] = value.ToString(); save(); }
        }

       
        /// <summary>
        /// Read the xml file, process nodes under keyword node as key/value pairs.
        /// </summary>
        /// <returns>null if no errors, else the error message</returns>
        /// 

        /*
         * The config file can be altered by the user and hence can contain errors: keyword not found or
         * misspelled, expected numeric data not numeric, config file not even present.  So, allow for
         * these errors by setting defaults and using a try/catch block:
         */
        /*The xconfig.xml file should look like the following (exactly, except for the keyword/value pairs:
         * 
            <?xml version="1.0" encoding="utf-8"?>
            <config>
                <keyword>
                    <minFreq>414</minFreq>
                    <maxFreq>828</maxFreq>
                </keyword>
             </config>
         */

        /*  Here is sample code that should appear in your program:
          
            int minFreq = 1440;    //You should set default values in case of failure below
            int maxFreq = 2880;
            
            //In your main program, at startup time
            Config config = new Config(initValuesDictionary); //assumes file is in executable directory and called xconfig.xml
            
         * 
            config = XConfig.theConfig;  //shorthand in your function for the static variable.
            
         * 
         *  You can save the config at any time.  It is also saved upon Dispose  (only do that once).
            config.save();
         * 
         * If you change a config value at some point, it is up to you to put it in the dictionary, as
                config["kword"] = newValue;
         * 
         */
        /// <summary>
        /// We save the xml config file by saving the dictionary in that form.  You can call save()
        /// at anytime yourself.
        /// </summary>

        public void Dispose()
        {
            if (theConfig != null)
                save();
            theConfig = null;
        }

        public static string GetString(string key)
        {
            return theConfig[key];
        } 

     
        private string init()
        {
                //If config file does not exist we will create it
                //If is is corrupt, it gets reset.
            XDocument aConfig = null; ;
            try
            {
                aConfig = XDocument.Load(fileName);
            
                IEnumerable<XNode> nodes =
                    from el in aConfig.Root.Elements("keyword").Nodes()
                    select el;
            
                foreach (XElement aNode in nodes)
                {
                    string aKey = aNode.Name.ToString();
                    string aValue = aNode.Value.ToString();

                    if (keywordConfig.ContainsKey(aKey))
                        keywordConfig[aKey] = aValue;
                    else keywordConfig.Add(aKey, aValue);
               
                }
            }
            catch (Exception)
            {
                reset();
                newFile = true;
            }

            return null;
        }

        public void reset()   //effectively will restore the defaults .
        {
            keywordConfig.Clear();
        }

        public void save()
        {
            XDocument doc = new XDocument(new XDeclaration("1.0", "UTF-8", "yes"),
               new System.Xml.Linq.XElement("config"));  

            doc.Root.Add(new XElement("keyword"));

            var configNode = doc.Descendants("keyword").FirstOrDefault();

            foreach (KeyValuePair<string, string> pair in keywordConfig)
            {
                XElement kwElement = new XElement(pair.Key, pair.Value);
                configNode.Add(kwElement);
            }

            try
            {
                StreamWriter sw;
                using (sw = new StreamWriter(fileName))
                {
                    doc.Save(sw);
                }
            }
            catch (Exception)
            {

            }
        }

        public static void Set(string key, object value)
        {
            XConfig.theConfig.keywordConfig[key] = value.ToString();
            XConfig.theConfig.save();   
        } 
    }
}
